"""Classes used to simulate recurly resources and endpoints

Each endpoint class will define the CRUD interface into the resource.
"""
import recurly
import six
import random
import string
import dateutil.relativedelta
import dateutil.parser

from .utils import current_time
from .errors import TRANSACTION_ERRORS, ResponseError
from .utils import details_route, serialize, serialize_list
from .backend import accounts_backend, billing_info_backend, transactions_backend, invoices_backend, subscriptions_backend, plans_backend, plan_add_ons_backend, adjustments_backend, coupons_backend, coupon_redemptions_backend


class BaseRecurlyEndpoint(object):
    """Baseclass for simulating resource endpoints.

    Provides basic CRUD functionality given a resource XML template, and object
    store backend.
    """
    pk_attr = 'uuid'
    XML = 0
    RAW = 1

    def hydrate_foreign_keys(self, obj):
        """Hydrates all foreign key objects from Id strings into actual objects
        """
        return obj

    def get_object_uri(self, obj):
        """Returns the URI to access the given object resource
        """
        cls = self.__class__
        return recurly.base_uri() + cls.base_uri + '/' + obj[cls.pk_attr]

    def uris(self, obj):
        """Returns a dictionary of all URIs related to the object, including foreign keys
        """
        obj = self.hydrate_foreign_keys(obj)
        uri_out = {}
        uri_out['object_uri'] = self.get_object_uri(obj)
        return uri_out

    def serialize(self, obj, format=XML):
        """Serialize the object into the provided format, using the resource
        template.

        Currently only supports XML (for XML representation of the resource.
        This is what recurly expects) and RAW (a dictionary representation of
        the resource)
        """
        if format == BaseRecurlyEndpoint.RAW:
            return obj

        cls = self.__class__
        if type(obj) == list:
            for o in obj:
                o['uris'] = self.uris(o)
            return serialize_list(cls.template, cls.object_type_plural, cls.object_type, obj)
        else:
            obj['uris'] = self.uris(obj)
            return serialize(cls.template, cls.object_type, obj)

    def list(self, format=XML):
        """Endpoint to list all resources stored in the backend
        """
        cls = self.__class__
        out = cls.backend.list_objects()
        return self.serialize(out, format=format)

    def create(self, create_info, format=XML):
        """Endpoint to create a new instance of the resource into the backend
        """
        cls = self.__class__
        if cls.pk_attr in create_info:
            create_info['uuid'] = create_info[cls.pk_attr]
        else:
            create_info['uuid'] = self.generate_id()
        new_obj = cls.backend.add_object(create_info['uuid'], create_info)
        return self.serialize(new_obj, format=format)

    def retrieve(self, pk, format=XML):
        """Endpoint to retrieve an existing resource from the backend

        Raises a 404 if the requested object does not exist.
        """
        cls = self.__class__
        if not cls.backend.has_object(pk):
            raise ResponseError(404, '')
        out = cls.backend.get_object(pk)
        return self.serialize(out, format=format)

    def update(self, pk, update_info, format=XML):
        """Endpoint to update an existing resource from the backend

        Raises a 404 if the requested object does not exist.
        """
        cls = self.__class__
        if not cls.backend.has_object(pk):
            raise ResponseError(404, '')
        out = cls.backend.update_object(pk, update_info)
        return self.serialize(out, format=format)

    def delete(self, pk):
        """Endpoint to delete an existing resource from the backend

        Raises a 404 if the requested object does not exist.
        """
        cls = self.__class__
        if not cls.backend.has_object(pk):
            raise ResponseError(404, '')
        cls.backend.delete_object(pk)
        return ''

    def generate_id(self):
        """Generates a random ID that can be used as a UUID or recurly ID
        """
        return ''.join(random.choice(string.ascii_lowercase + string.digits) for i in range(32))


class AccountsEndpoint(BaseRecurlyEndpoint):
    base_uri = 'accounts'
    pk_attr = 'account_code'
    backend = accounts_backend
    object_type = 'account'
    object_type_plural = 'accounts'
    template = 'account.xml'

    def uris(self, obj):
        uri_out = super(AccountsEndpoint, self).uris(obj)
        uri_out['adjustments_uri'] = uri_out['object_uri'] + '/adjustments'
        if billing_info_backend.has_object(obj[AccountsEndpoint.pk_attr]):
            uri_out['billing_info_uri'] = uri_out['object_uri'] + '/billing_info'
        uri_out['invoices_uri'] = uri_out['object_uri'] + '/invoices'
        uri_out['redemption_uri'] = uri_out['object_uri'] + '/redemption'
        uri_out['subscriptions_uri'] = uri_out['object_uri'] + '/subscriptions'
        uri_out['transactions_uri'] = uri_out['object_uri'] + '/transactions'
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        if 'billing_info' in create_info:
            billing_info = create_info['billing_info']
            billing_info['account'] = create_info['account_code']
            billing_info_backend.add_object(create_info[AccountsEndpoint.pk_attr], billing_info)
            del create_info['billing_info']
        create_info['hosted_login_token'] = self.generate_id()
        create_info['created_at'] = current_time().isoformat()
        return super(AccountsEndpoint, self).create(create_info, format=format)

    def update(self, pk, update_info, format=BaseRecurlyEndpoint.XML):
        if 'billing_info' in update_info:
            updated_billing_info = update_info['billing_info']
            if billing_info_backend.has_object(pk):
                billing_info_backend.update_object(pk, updated_billing_info)
            else:
                updated_billing_info['account'] = pk
                billing_info_backend.add_object(pk, updated_billing_info)
            del update_info['billing_info']
        return super(AccountsEndpoint, self).update(pk, update_info, format=format)

    def delete(self, pk):
        AccountsEndpoint.backend.update_object(pk, {'state': 'closed'})
        billing_info_backend.delete_object(pk)
        return ''

    # Support for nested resources
    # BillingInfo and CouponRedemption are managed by this endpoint, as
    # opposed to having their own since Recurly API only provides access to these
    # resources through the Account endpoint.
    def billing_info_uris(self, obj):
        uri_out = {}
        uri_out['account_uri'] = recurly.base_uri() + AccountsEndpoint.base_uri + '/' + obj['account']
        uri_out['object_uri'] = uri_out['account_uri'] + '/billing_info'
        return uri_out

    def serialize_billing_info(self, obj, format=BaseRecurlyEndpoint.XML):
        if format == BaseRecurlyEndpoint.RAW:
            return obj

        obj['uris'] = self.billing_info_uris(obj)
        return serialize('billing_info.xml', 'billing_info', obj)

    @details_route('GET', 'billing_info')
    def get_billing_info(self, pk, format=BaseRecurlyEndpoint.XML):
        out = billing_info_backend.get_object(pk)
        return self.serialize_billing_info(out, format=format)

    @details_route('PUT', 'billing_info')
    def update_billing_info(self, pk, update_info, format=BaseRecurlyEndpoint.XML):
        if billing_info_backend.has_object(pk):
            out = billing_info_backend.update_object(pk, update_info)
        else:
            update_info['account'] = self.pk_attr
            out = billing_info_backend.add_object(pk, update_info)
        return self.serialize_billing_info(out, format=format)

    @details_route('DELETE', 'billing_info')
    def delete_billing_info(self, pk):
        billing_info_backend.delete_object(pk)
        return ''

    @details_route('GET', 'transactions', is_list=True)
    def get_transactions_list(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
        out = TransactionsEndpoint.backend.list_objects(lambda transaction: transaction['account'] == pk)
        return transactions_endpoint.serialize(out, format=format)

    @details_route('GET', 'invoices', is_list=True)
    def get_invoices_list(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
        out = InvoicesEndpoint.backend.list_objects(lambda invoice: invoice['account'] == pk)
        return invoices_endpoint.serialize(out, format=format)

    @details_route('GET', 'subscriptions', is_list=True)
    def get_subscriptions_list(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
        def filter_subscriptions(subscription):
            if filters:
                if 'state' in filters and filters['state'][0] == 'live':
                    filters['state'] = ['active', 'canceled', 'future', 'in_trial']
                cond = all(subscription[k] in v for k, v in filters.items())
            else:
                cond = True
            return subscription['account'] == pk and cond
        out = SubscriptionsEndpoint.backend.list_objects(filter_subscriptions)
        return subscriptions_endpoint.serialize(out, format=format)

    @details_route('GET', 'redemption')
    def get_coupon_redemption_view(self, account_code, format=BaseRecurlyEndpoint.XML):
        coupon_redemption = self.get_coupon_redemption(account_code)
        if coupon_redemption is None:
            raise ResponseError(404, '')
        return coupons_endpoint.serialize_coupon_redemption(coupon_redemption, format=format)

    def get_coupon_redemption(self, account_code):
        account_coupon_redemptions = coupon_redemptions_backend.list_objects(lambda redemption: redemption['account_code'] == account_code)
        if len(account_coupon_redemptions) == 0:
            return None

        assert len(account_coupon_redemptions) == 1
        coupon_redemption = account_coupon_redemptions[0]
        return coupons_endpoint.hydrate_coupon_redemption_foreign_keys(coupon_redemption)

    @details_route('DELETE', 'redemption')
    def delete_coupon_redemption(self, account_code, format=BaseRecurlyEndpoint.XML):
        coupon_redemption = self.get_coupon_redemption(account_code)
        if coupon_redemption is None:
            raise ResponseError(404, '')

        coupon_redemption_uuid = coupons_endpoint.generate_coupon_redemption_uuid(coupon_redemption['coupon']['coupon_code'], account_code)
        coupon_redemptions_backend.delete_object(coupon_redemption_uuid)
        return ''


class TransactionsEndpoint(BaseRecurlyEndpoint):
    base_uri = 'transactions'
    backend = transactions_backend
    object_type = 'transaction'
    object_type_plural = 'transactions'
    template = 'transaction.xml'

    def __init__(self):
        self.registered_errors = {}
        return super(TransactionsEndpoint, self).__init__()

    def clear_state(self):
        """Clears all registered errors
        """
        self.registered_errors = {}

    def register_transaction_failure(self, account_code, error_code):
        """Registers an error_code to associate with the given account for all
        transactions made by the account
        """
        self.registered_errors[account_code] = error_code

    def hydrate_foreign_keys(self, obj):
        if isinstance(obj['account'], six.string_types):
            # hydrate account
            obj['account'] = AccountsEndpoint.backend.get_object(obj['account'])
        if 'invoice' in obj and isinstance(obj['invoice'], six.string_types):
            # hydrate invoice
            obj['invoice'] = InvoicesEndpoint.backend.get_object(obj['invoice'])
        return obj

    def uris(self, obj):
        uri_out = super(TransactionsEndpoint, self).uris(obj)
        obj['account']['uris'] = accounts_endpoint.uris(obj['account'])
        uri_out['account_uri'] = obj['account']['uris']['object_uri']
        if 'invoice' in obj:
            # To avoid infinite recursion
            uri_out['invoice_uri'] = invoices_endpoint.get_object_uri(obj['invoice'])
        if 'subscription' in obj:
            pseudo_subscription_object = {}
            pseudo_subscription_object[SubscriptionsEndpoint.pk_attr] = obj['subscription']
            uri_out['subscription_uri'] = subscriptions_endpoint.get_object_uri(pseudo_subscription_object)
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        # Like recurly, creates an invoice that is associated with the
        # transaction
        account_code = create_info['account'][AccountsEndpoint.pk_attr]
        assert AccountsEndpoint.backend.has_object(account_code)
        create_info['account'] = account_code

        create_info['uuid'] = self.generate_id()  # generate id now for invoice
        create_info['tax_in_cents'] = 0  # unsupported
        create_info['action'] = 'purchase'
        create_info['status'] = 'success'
        create_info['test'] = True
        create_info['voidable'] = True
        create_info['refundable'] = True
        create_info['created_at'] = current_time().isoformat()
        create_info['type'] = 'credit_card'
        if 'description' not in create_info:
            create_info['description'] = ''

        # Check to see if we need to throw an error for card failure
        if create_info['account'] in self.registered_errors:
            # update the new transaction with error info
            create_info['voidable'] = False
            create_info['refundable'] = False
            create_info['status'] = 'declined'
            error_code = self.registered_errors[create_info['account']]
            transaction_error = TRANSACTION_ERRORS[error_code]
            create_info['transaction_error'] = transaction_error
            transaction_xml = super(TransactionsEndpoint, self).create(create_info, format)
            error_xml = serialize('transaction_error.xml', 'transaction_error', transaction_error)
            raise ResponseError(422, '<errors>{0}{1}</errors>'.format(error_xml, transaction_xml))

        # Every new transaction creates a new invoice
        new_invoice = {'account': account_code,
                       'uuid': self.generate_id(),
                       'state': 'collected',
                       'invoice_number': InvoicesEndpoint.generate_invoice_number(),
                       'subtotal_in_cents': int(create_info['amount_in_cents']),
                       'currency': create_info['currency'],
                       'created_at': create_info['created_at'],
                       'net_terms': 0,
                       'collection_method': 'automatic',

                       # unsupported
                       'tax_type': 'usst',
                       'tax_rate': 0}
        new_invoice['tax_in_cents'] = new_invoice['subtotal_in_cents'] * new_invoice['tax_rate']
        new_invoice['total_in_cents'] = new_invoice['subtotal_in_cents'] + new_invoice['tax_in_cents']
        new_invoice['transactions'] = [create_info['uuid']]
        InvoicesEndpoint.backend.add_object(new_invoice['invoice_number'], new_invoice)
        new_invoice_id = new_invoice[InvoicesEndpoint.pk_attr]

        # Every transaction should have a line item as well
        transaction_charge_line_item = {'account_code': new_invoice['account'],
                                        'currency': new_invoice['currency'],
                                        'unit_amount_in_cents': int(new_invoice['total_in_cents']),
                                        'description': create_info['description'],
                                        'quantity': 1,
                                        'invoice': new_invoice_id}
        transaction_charge_line_item = adjustments_endpoint.create(transaction_charge_line_item, format=BaseRecurlyEndpoint.RAW)
        InvoicesEndpoint.backend.update_object(new_invoice_id, {'line_items': [transaction_charge_line_item]})

        create_info['invoice'] = new_invoice_id
        return super(TransactionsEndpoint, self).create(create_info, format)

    def delete(self, pk, amount_in_cents=None):
        """As of Nov. 2014, DELETE on transactions is no longer implemented
        """
        raise ResponseError(404, '')


class AdjustmentsEndpoint(BaseRecurlyEndpoint):
    base_uri = 'adjustments'
    backend = adjustments_backend
    object_type = 'adjustment'
    object_type_plural = 'adjustments'
    template = 'adjustment.xml'
    defaults = {'state': 'active',
                'quantity': 1,
                'origin': 'credit',
                'product_code': 'basic',
                'discount_in_cents': 0,
                # unsupported
                'tax_exempt': False}

    def uris(self, obj):
        uri_out = super(AdjustmentsEndpoint, self).uris(obj)
        pseudo_account_object = {}
        pseudo_account_object[AccountsEndpoint.pk_attr] = obj['account_code']
        uri_out['account_uri'] = accounts_endpoint.get_object_uri(pseudo_account_object)
        pseudo_invoice_object = {}
        pseudo_invoice_object[InvoicesEndpoint.pk_attr] = obj['invoice']
        uri_out['invoice_uri'] = invoices_endpoint.get_object_uri(pseudo_invoice_object)
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        create_info['created_at'] = create_info['start_date'] = current_time().isoformat()
        if int(create_info['unit_amount_in_cents']) >= 0:
            create_info['type'] = 'charge'
        else:
            create_info['type'] = 'credit'

        # UNSUPPORTED
        create_info['tax_in_cents'] = 0
        create_info['total_in_cents'] = int(create_info['unit_amount_in_cents']) + int(create_info['tax_in_cents'])

        defaults = AdjustmentsEndpoint.defaults.copy()
        defaults.update(create_info)
        defaults['total_in_cents'] -= defaults['discount_in_cents']

        return super(AdjustmentsEndpoint, self).create(defaults, format)


class InvoicesEndpoint(BaseRecurlyEndpoint):
    base_uri = 'invoices'
    backend = invoices_backend
    object_type = 'invoice'
    object_type_plural = 'invoices'
    pk_attr = 'invoice_number'
    template = 'invoice.xml'

    def hydrate_foreign_keys(self, obj):
        if isinstance(obj['account'], six.string_types):
            # hydrate account
            obj['account'] = AccountsEndpoint.backend.get_object(obj['account'])
        if 'transactions' in obj:
            obj['transactions'] = [TransactionsEndpoint.backend.get_object(transaction_id) if isinstance(transaction_id, six.string_types) else transaction_id for transaction_id in obj['transactions']]
            for transaction in obj['transactions']:
                transaction['invoice'] = obj
                transaction['uris'] = transactions_endpoint.uris(transaction)
        if 'line_items' in obj:
            obj['line_items'] = [AdjustmentsEndpoint.backend.get_object(adjustment_id) if isinstance(adjustment_id, six.string_types) else adjustment_id for adjustment_id in obj['line_items']]
            for adjustment in obj['line_items']:
                adjustment['uris'] = adjustments_endpoint.uris(adjustment)
        return obj

    def uris(self, obj):
        uri_out = super(InvoicesEndpoint, self).uris(obj)
        uri_out['account_uri'] = accounts_endpoint.get_object_uri(obj['account'])
        if 'subscription' in obj:
            uri_out['subscription_uri'] = subscriptions_endpoint.get_object_uri({'uuid': obj['subscription']})
        if 'original_invoice' in obj:
            faux_inv = {}
            faux_inv[InvoicesEndpoint.pk_attr] = obj['original_invoice']
            uri_out['original_invoice_uri'] = self.get_object_uri(faux_inv)
        return uri_out

    @details_route('POST', 'refund')
    def refund_invoice(self, pk, refund_info, format=BaseRecurlyEndpoint.XML):
        """Refunds the invoice.

        There are two ways this can happen:
        - Refund individual line items on the invoice
        - Refund a specific amount

        The outcome will:
        - Create a new invoice with adjustments that cancel out the original
          invoice (invoice_number = pk)
        - Updates the state of associated objects
        """
        invoice = InvoicesEndpoint.backend.get_object(pk)

        if 'amount_in_cents' in refund_info:
            return self._refund_amount(invoice, int(refund_info['amount_in_cents']))
        else:
            # Hack to get around the singleton hydration of XML
            if isinstance(refund_info['line_items'], dict):
                refund_info['line_items'] = [refund_info['line_items']]
            return self._refund_line_items(invoice, refund_info)

    def _refund_amount(self, invoice, amount_in_cents):
        """Refunds a specific amount for the invoice."""

        # Create a new transaction that tracks the refund
        refund_transaction_info = {
            'account': accounts_endpoint.backend.get_object(invoice['account']),
            'amount_in_cents': -amount_in_cents,
            'currency': 'USD',
            'description': 'Refund for Invoice #{}'.format(invoice['invoice_number'])
        }
        new_transaction = transactions_endpoint.create(refund_transaction_info, format=BaseRecurlyEndpoint.RAW)

        # Update transaction to mimic refund transaction
        opts = {
            'action': 'refund',
            'refundable': False,
        }
        if 'subscription' in invoice:
            opts['subscription'] = invoice['subscription']
        TransactionsEndpoint.backend.update_object(new_transaction['uuid'], opts)

        # Update adjustments to mimic refund invoice
        new_transaction = TransactionsEndpoint.backend.get_object(new_transaction['uuid'])
        new_invoice = InvoicesEndpoint.backend.get_object(new_transaction['invoice'])
        adjustments = new_invoice['line_items']
        new_adjustments = []
        for adjustment in adjustments:
            new_adjustments.append(AdjustmentsEndpoint.backend.update_object(adjustment['uuid'], {'quantity': -adjustment['quantity']}))
        new_invoice = InvoicesEndpoint.backend.update_object(new_invoice['invoice_number'], {'line_items': new_adjustments, 'original_invoice': invoice[InvoicesEndpoint.pk_attr]})

        return self.serialize(new_invoice)

    def _refund_line_items(self, invoice, refund_info):
        """Refund individual line items on the invoice."""

        # Create the refund line items
        refund_line_items = self._create_refund_line_items_for(refund_info['line_items'])

        # Calculate amount to refund
        amount_to_refund = -sum(map(lambda line_item: line_item['unit_amount_in_cents'], refund_line_items))

        # New invoice tracking refund
        new_invoice = self._create_refund_invoice_for(invoice, amount_to_refund)
        new_invoice_id = new_invoice[InvoicesEndpoint.pk_attr]

        # Relate the objects
        refund_line_items = map(lambda line_item: AdjustmentsEndpoint.backend.update_object(line_item[AdjustmentsEndpoint.pk_attr], {'invoice': new_invoice[InvoicesEndpoint.pk_attr]}), refund_line_items)
        new_invoice = InvoicesEndpoint.backend.update_object(new_invoice_id, {'line_items': refund_line_items})

        # Update transactions
        transactions = map(lambda t_pk: TransactionsEndpoint.backend.get_object(t_pk), invoice['transactions'])
        transactions_to_add = self._update_or_create_refund_transactions_for(transactions, new_invoice)
        new_invoice = InvoicesEndpoint.backend.update_object(new_invoice_id, {'transactions': transactions_to_add})

        return self.serialize(new_invoice)

    def _create_refund_line_items_for(self, line_items):
        """Creates refund line items for the given line items.

        Returns any new line items that were created
        """
        refund_line_items = []
        for line_item in line_items:
            quantity_to_refund = line_item['adjustment']['quantity']
            # TODO: add logic for prorate
            line_item = AdjustmentsEndpoint.backend.get_object(line_item['adjustment']['uuid'])

            assert int(quantity_to_refund) <= int(line_item['quantity'])

            charge_refund_line_item = {'account_code': line_item['account_code'],
                                       'currency': line_item['currency'],
                                       'unit_amount_in_cents': -int(line_item['unit_amount_in_cents']),
                                       'description': 'Refund for {}'.format(line_item['description']),
                                       'quantity': -int(quantity_to_refund)}
            charge_refund_line_item = adjustments_endpoint.create(charge_refund_line_item, format=BaseRecurlyEndpoint.RAW)
            refund_line_items.append(charge_refund_line_item)
        return refund_line_items

    def _create_refund_invoice_for(self, invoice, amount_to_refund):
        """Creates the refund invoice for the given invoice."""
        # New invoice tracking refund
        new_invoice = {'account': invoice['account'],
                       'uuid': self.generate_id(),
                       'state': 'collected',
                       'invoice_number': InvoicesEndpoint.generate_invoice_number(),
                       'subtotal_in_cents': -int(amount_to_refund),
                       'currency': invoice['currency'],
                       'created_at': current_time().isoformat(),
                       'net_terms': 0,
                       'collection_method': 'automatic',
                       'original_invoice': invoice[InvoicesEndpoint.pk_attr],

                       # unsupported
                       'tax_type': 'usst',
                       'tax_rate': 0}
        new_invoice['tax_in_cents'] = new_invoice['subtotal_in_cents'] * new_invoice['tax_rate']
        new_invoice['total_in_cents'] = new_invoice['subtotal_in_cents'] + new_invoice['tax_in_cents']
        new_invoice = InvoicesEndpoint.backend.add_object(new_invoice[InvoicesEndpoint.pk_attr], new_invoice)
        return new_invoice

    def _update_or_create_refund_transactions_for(self, transactions, new_invoice):
        """
        Updates existing transactions to be void (if voidable), or creates a
        new transaction to track refund.
        """
        # Update state of any associated objects
        #   If invoice is with transaction, then void/refund the transaction
        transactions_to_add = []
        for transaction in transactions:
            if transaction['voidable']:
                TransactionsEndpoint.backend.update_object(transaction['uuid'], {
                    'status': 'void',
                    'voidable': False,
                    'refundable': False  # TODO: only for full refunds
                })
                transactions_to_add.append(transaction['uuid'])
            else:
                new_transaction = {
                    'uuid': transactions_endpoint.generate_id(),
                    'action': 'refund',
                    'status': 'success',
                    'test': True,
                    'voidable': True,
                    'refundable': False,
                    'created_at': current_time().isoformat(),
                    'type': 'credit_card',
                    'account': new_invoice['account'],
                    'currency': new_invoice['currency'],
                    'amount_in_cents': int(new_invoice['total_in_cents']),
                    'invoice': new_invoice[InvoicesEndpoint.pk_attr],

                    # unsupported
                    'tax_in_cents': 0
                }
                TransactionsEndpoint.backend.add_object(new_transaction['uuid'], new_transaction)
                transactions_to_add.append(new_transaction['uuid'])
                TransactionsEndpoint.backend.update_object(transaction['uuid'], {'refundable': False})
        return transactions_to_add

    @staticmethod
    def generate_invoice_number():
        if InvoicesEndpoint.backend.empty():
            return '1000'
        return str(max(int(invoice['invoice_number']) for invoice in InvoicesEndpoint.backend.list_objects()) + 1)


class CouponsEndpoint(BaseRecurlyEndpoint):
    base_uri = 'coupons'
    backend = coupons_backend
    object_type = 'coupon'
    object_type_plural = 'coupons'
    pk_attr = 'coupon_code'
    template = 'coupon.xml'
    defaults = {'state': 'redeemable',
                'applies_to_all_plans': True,
                'single_use': False}

    def uris(self, obj):
        uri_out = super(CouponsEndpoint, self).uris(obj)
        uri_out['redemptions_uri'] = uri_out['object_uri'] + '/redemptions'
        uri_out['redeem_uri'] = uri_out['object_uri'] + '/redeem'
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        defaults = CouponsEndpoint.defaults.copy()
        defaults.update(create_info)
        return super(CouponsEndpoint, self).create(defaults, format)

    def generate_coupon_redemption_uuid(self, coupon_code, account_code):
        return '__'.join([coupon_code, account_code])

    def hydrate_coupon_redemption_foreign_keys(self, obj):
        if isinstance(obj['coupon'], six.string_types):
            obj['coupon'] = CouponsEndpoint.backend.get_object(obj['coupon'])
        return obj

    def coupon_redemption_uris(self, obj):
        uri_out = {}
        uri_out['coupon_uri'] = coupons_endpoint.get_object_uri(obj['coupon'])
        pseudo_account_object = {}
        pseudo_account_object[AccountsEndpoint.pk_attr] = obj['account_code']
        uri_out['account_uri'] = accounts_endpoint.get_object_uri(pseudo_account_object)
        uri_out['object_uri'] = uri_out['account_uri'] + '/redemption'
        return uri_out

    def serialize_coupon_redemption(self, obj, format=BaseRecurlyEndpoint.XML):
        if format == BaseRecurlyEndpoint.RAW:
            obj = self.hydrate_coupon_redemption_foreign_keys(obj)
            return obj

        if type(obj) == list:
            obj = [self.hydrate_coupon_redemption_foreign_keys(o) for o in obj]
            for o in obj:
                o['uris'] = self.coupon_redemption_uris(o)
            return serialize_list('redemption.xml', 'redemptions', 'redemption', obj)
        else:
            obj = self.hydrate_coupon_redemption_foreign_keys(obj)
            obj['uris'] = self.coupon_redemption_uris(obj)
            return serialize('redemption.xml', 'redemption', obj)

    @details_route('GET', 'redemptions', is_list=True)
    def get_coupon_redemptions(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
        obj_list = coupon_redemptions_backend.list_objects(lambda redemption: redemption['coupon'] == pk)
        return self.serialize_coupon_redemption(obj_list, format=format)

    @details_route('POST', 'redeem')
    def redeem_coupon(self, pk, redeem_info, format=BaseRecurlyEndpoint.XML):
        assert CouponsEndpoint.backend.has_object(pk)
        redeem_info['coupon'] = pk
        redeem_info['created_at'] = current_time().isoformat()
        return self.serialize_coupon_redemption(coupon_redemptions_backend.add_object(self.generate_coupon_redemption_uuid(pk, redeem_info['account_code']), redeem_info), format=format)

    def determine_coupon_discount(self, coupon, charge):
        type = coupon['discount_type']
        if type == 'percent':
            return int(charge * float(coupon['discount_percent']) / 100)
        else:
            return int(coupon['discount_in_cents'])


class PlansEndpoint(BaseRecurlyEndpoint):
    base_uri = 'plans'
    backend = plans_backend
    pk_attr = 'plan_code'
    object_type = 'plan'
    object_type_plural = 'plans'
    template = 'plan.xml'
    defaults = {'plan_interval_unit': 'months',
                'plan_interval_length': 1,
                'trial_interval_unit': 'months',
                'trial_interval_length': 0,
                'display_quantity': False,
                # unsupported
                'tax_exempt': False}
    add_on_defaults = {'default_quantity': 1,
                       'display_quantity_on_hosted_page': False}

    def uris(self, obj):
        uri_out = super(PlansEndpoint, self).uris(obj)
        uri_out['add_ons_uri'] = uri_out['object_uri'] + '/add_ons'
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        create_info['created_at'] = current_time().isoformat()
        defaults = PlansEndpoint.defaults.copy()
        defaults.update(create_info)
        return super(PlansEndpoint, self).create(defaults, format)

    def generate_plan_add_on_uuid(self, plan_code, add_on_code):
        return '__'.join([plan_code, add_on_code])

    def plan_add_on_uris(self, obj):
        uri_out = {}
        pseudo_plan_object = {}
        pseudo_plan_object[PlansEndpoint.pk_attr] = obj['plan']
        uri_out['plan_uri'] = plans_endpoint.get_object_uri(pseudo_plan_object)
        uri_out['object_uri'] = uri_out['plan_uri'] + '/add_ons/' + obj['add_on_code']
        return uri_out

    def serialize_plan_add_on(self, obj, format=BaseRecurlyEndpoint.XML):
        if format == BaseRecurlyEndpoint.RAW:
            return obj

        if type(obj) == list:
            for o in obj:
                o['uris'] = self.plan_add_on_uris(o)
            return serialize_list('add_on.xml', 'add_ons', 'add_on', obj)
        else:
            obj['uris'] = self.plan_add_on_uris(obj)
            return serialize('add_on.xml', 'add_on', obj)

    @details_route('GET', 'add_ons', is_list=True)
    def get_add_on_list(self, pk, filters=None, format=BaseRecurlyEndpoint.XML):
        out = plan_add_ons_backend.list_objects(lambda add_on: add_on['plan'] == pk)
        return self.serialize_plan_add_on(out, format=format)

    @details_route('POST', 'add_ons')
    def create_add_on(self, pk, create_info, format=BaseRecurlyEndpoint.XML):
        assert PlansEndpoint.backend.has_object(pk)
        create_info['plan'] = pk
        create_info['created_at'] = current_time().isoformat()
        if 'accounting_code' not in create_info:
            create_info['accounting_code'] = create_info['add_on_code']
        return self.serialize_plan_add_on(plan_add_ons_backend.add_object(self.generate_plan_add_on_uuid(pk, create_info['add_on_code']), create_info), format=format)


class SubscriptionsEndpoint(BaseRecurlyEndpoint):
    base_uri = 'subscriptions'
    backend = subscriptions_backend
    object_type = 'subscription'
    object_type_plural = 'subscriptions'
    template = 'subscription.xml'
    defaults = {'quantity': 1, 'collection_method': 'automatic'}

    def _calculate_timedelta(self, units, length):
        timedelta_info = {}
        timedelta_info[units] = int(length)
        return dateutil.relativedelta.relativedelta(**timedelta_info)

    def _parse_isoformat(self, isoformat):
        return dateutil.parser.parse(isoformat)

    def hydrate_foreign_keys(self, obj):
        if 'plan' not in obj:
            obj['plan'] = PlansEndpoint.backend.get_object(obj['plan_code'])
        if 'subscription_add_ons' in obj:
            def hydrate_add_ons(add_on):
                if isinstance(add_on, six.string_types):
                    add_on = plan_add_ons_backend.get_object(plans_endpoint.generate_plan_add_on_uuid(obj['plan_code'], add_on))
                    add_on['unit_amount_in_cents'] = add_on['unit_amount_in_cents'][obj['currency']]
                return add_on
            obj['subscription_add_ons'] = map(hydrate_add_ons, obj['subscription_add_ons'])
        return obj

    def uris(self, obj):
        uri_out = super(SubscriptionsEndpoint, self).uris(obj)
        pseudo_account_object = {}
        pseudo_account_object[AccountsEndpoint.pk_attr] = obj['account']
        uri_out['account_uri'] = accounts_endpoint.get_object_uri(pseudo_account_object)
        if 'invoice' in obj:
            pseudo_invoice_object = {}
            pseudo_invoice_object[InvoicesEndpoint.pk_attr] = obj['invoice']
            uri_out['invoice_uri'] = invoices_endpoint.get_object_uri(pseudo_invoice_object)
        uri_out['plan_uri'] = plans_endpoint.get_object_uri(obj['plan'])
        uri_out['cancel_uri'] = uri_out['object_uri'] + '/cancel'
        uri_out['reactivate_uri'] = uri_out['object_uri'] + '/reactivate'
        uri_out['terminate_uri'] = uri_out['object_uri'] + '/terminate'
        return uri_out

    def create(self, create_info, format=BaseRecurlyEndpoint.XML):
        # Like recurly, this will create a new invoice and transaction that
        # goes with the new subscription enrollment
        account_code = create_info['account'][AccountsEndpoint.pk_attr]
        if not AccountsEndpoint.backend.has_object(account_code):
            accounts_endpoint.create(create_info['account'])
        else:
            accounts_endpoint.update(account_code, create_info['account'])
        create_info['account'] = account_code

        assert plans_backend.has_object(create_info['plan_code'])
        plan = plans_backend.get_object(create_info['plan_code'])

        now = current_time()

        # Trial dates need to be calculated
        if 'trial_ends_at' in create_info:
            create_info['trial_started_at'] = now.isoformat()
        elif plan['trial_interval_length'] > 0:
            create_info['trial_started_at'] = now.isoformat()
            create_info['trial_ends_at'] = (now + self._calculate_timedelta(plan['trial_interval_unit'], plan['trial_interval_length'])).isoformat()

        # Plan start and end date needs to be calculated
        if 'starts_at' in create_info:
            # A custom start date is specified
            create_info['activated_at'] = create_info['starts_at']
            # TODO: confirm recurly sets current_period_started_at for future subs
            create_info['current_period_started_at'] = create_info['starts_at']
        elif 'trial_started_at' in create_info:
            create_info['activated_at'] = self._parse_isoformat(create_info['trial_ends_at'])
            create_info['current_period_started_at'] = create_info['trial_started_at']
            create_info['current_period_ends_at'] = create_info['trial_ends_at']
        else:
            create_info['activated_at'] = now.isoformat()
            create_info['current_period_started_at'] = now.isoformat()

        started_at = self._parse_isoformat(create_info['current_period_started_at'])
        if now >= started_at:
            # Plan already started
            if 'first_renewal_date' in create_info:
                create_info['current_period_ends_at'] = self._parse_isoformat(create_info['first_renewal_date'])
            else:
                create_info['current_period_ends_at'] = (started_at + self._calculate_timedelta(plan['plan_interval_unit'], plan['plan_interval_length'])).isoformat()

        # Tax calculated based on plan info
        # UNSUPPORTED
        create_info['tax_in_cents'] = 0
        create_info['tax_type'] = 'usst'
        create_info['tax_rate'] = 0

        # Subscription states
        if 'trial_started_at' in create_info:
            create_info['state'] = 'trial'
        elif 'current_period_ends_at' not in create_info:
            create_info['state'] = 'future'
        else:
            create_info['state'] = 'active'

        # If there are addons, make sure they exist in the system
        if 'subscription_add_ons' in create_info:
            for add_on in create_info['subscription_add_ons']:
                add_on_uuid = plans_endpoint.generate_plan_add_on_uuid(create_info['plan_code'], add_on['add_on_code'])
                assert plan_add_ons_backend.has_object(add_on_uuid)
            create_info['subscription_add_ons'] = [add_on['add_on_code'] for add_on in create_info['subscription_add_ons']]

        defaults = SubscriptionsEndpoint.defaults.copy()
        defaults['unit_amount_in_cents'] = plan['unit_amount_in_cents'][create_info['currency']]
        defaults.update(create_info)

        # TODO: support bulk

        new_sub = super(SubscriptionsEndpoint, self).create(defaults, format=BaseRecurlyEndpoint.RAW)
        self.hydrate_foreign_keys(new_sub)

        if defaults['state'] == 'active':
            # Setup charges first, to calculate total charge to put on the
            # invoice and transaction
            total = 0
            adjustment_infos = []
            plan_charge_line_item = {
                'account_code': new_sub['account'],
                'currency': new_sub['currency'],
                'unit_amount_in_cents': int(new_sub['unit_amount_in_cents']),
                'description': new_sub['plan']['name'],
                'quantity': new_sub['quantity']
            }
            total += plan_charge_line_item['unit_amount_in_cents']
            adjustment_infos.append(plan_charge_line_item)

            if 'subscription_add_ons' in new_sub:
                for add_on in new_sub['subscription_add_ons']:
                    plan_charge_line_item = {
                        'account_code': new_sub['account'],
                        'currency': new_sub['currency'],
                        'unit_amount_in_cents': int(add_on['unit_amount_in_cents']),
                        'description': add_on['name'],
                        'quantity': new_sub['quantity'],
                    }
                    total += plan_charge_line_item['unit_amount_in_cents']
                    adjustment_infos.append(plan_charge_line_item)

            # now calculate discounts
            coupon_redemption = accounts_endpoint.get_coupon_redemption(new_sub['account'])
            if coupon_redemption:
                for plan_charge_line_item in adjustment_infos:
                    discount = coupons_endpoint.determine_coupon_discount(coupon_redemption['coupon'], plan_charge_line_item['unit_amount_in_cents'])
                    plan_charge_line_item['discount_in_cents'] = discount
                    total -= plan_charge_line_item['discount_in_cents']

            # create a transaction if the subscription is started
            new_transaction = {}
            new_transaction['account'] = {}
            new_transaction['account'][AccountsEndpoint.pk_attr] = new_sub['account']
            new_transaction['amount_in_cents'] = total
            new_transaction['currency'] = new_sub['currency']
            new_transaction['subscription'] = new_sub[SubscriptionsEndpoint.pk_attr]
            new_transaction = transactions_endpoint.create(new_transaction, format=BaseRecurlyEndpoint.RAW)
            new_invoice_id = new_transaction['invoice']

            # Now create accumulated new adjustments for the sub to track line items
            adjustments = []
            for plan_charge_line_item in adjustment_infos:
                plan_charge_line_item['invoice'] = new_invoice_id
                plan_charge_line_item = adjustments_endpoint.create(plan_charge_line_item, format=BaseRecurlyEndpoint.RAW)
                adjustments.append(plan_charge_line_item[AdjustmentsEndpoint.pk_attr])

            InvoicesEndpoint.backend.update_object(new_invoice_id, {'subscription': new_sub[SubscriptionsEndpoint.pk_attr], 'line_items': adjustments})

            new_sub = SubscriptionsEndpoint.backend.update_object(defaults['uuid'], {'invoice': new_invoice_id})
        return self.serialize(new_sub, format=format)

    @details_route('PUT', 'terminate')
    def terminate_subscription(self, pk, terminate_info, format=format):
        subscription = SubscriptionsEndpoint.backend.get_object(pk)
        # assume base transaction exists
        transaction = TransactionsEndpoint.backend.list_objects(lambda trans: trans['subscription'] == subscription[SubscriptionsEndpoint.pk_attr])[0]
        invoice_number = transaction['invoice']
        invoice = InvoicesEndpoint.backend.get_object(invoice_number)
        start = self._parse_isoformat(subscription['current_period_started_at'])
        end = self._parse_isoformat(subscription['current_period_ends_at'])
        now = current_time()
        refund_type = terminate_info['refund'][0]
        if refund_type == 'partial':
            if now > end:
                now = end
            days_left = (end - now).days
            total_days = (end - start).days
            refund_amount = int((float(days_left) / total_days) * transaction['amount_in_cents'])
            invoice_number = transaction['invoice']
            invoices_endpoint.refund_invoice(invoice_number, {'amount_in_cents': refund_amount})
        elif refund_type == 'full':
            adjustments_to_refund = []
            for line_item in invoice['line_items']:
                adjustments_to_refund.append({
                    'adjustment': AdjustmentsEndpoint.backend.get_object(line_item)
                })
            invoices_endpoint.refund_invoice(invoice_number, {'line_items': adjustments_to_refund})

        return self.serialize(SubscriptionsEndpoint.backend.update_object(pk, {
            'state': 'expired',
            'expires_at': now.isoformat(),
            'current_period_ends_at': now.isoformat()
        }), format=format)

    @details_route('PUT', 'cancel')
    def cancel_subscription(self, pk, cancel_info, format=format):
        subscription = SubscriptionsEndpoint.backend.get_object(pk)
        return self.serialize(SubscriptionsEndpoint.backend.update_object(pk, {
            'state': 'canceled',
            'expires_at': subscription['current_period_ends_at'],
            'canceled_at': current_time().isoformat()
        }), format=format)

    @details_route('PUT', 'reactivate')
    def reactivate_subscription(self, pk, reactivate_info, format=format):
        subscription = SubscriptionsEndpoint.backend.get_object(pk)
        if not subscription['state'] == 'canceled':
            raise ResponseError(400, '')
        return self.serialize(SubscriptionsEndpoint.backend.update_object(pk, {
            'state': 'active',
            'expires_at': None,
            'canceled_at': None
        }), format=format)

accounts_endpoint = AccountsEndpoint()
adjustments_endpoint = AdjustmentsEndpoint()
transactions_endpoint = TransactionsEndpoint()
coupons_endpoint = CouponsEndpoint()
invoices_endpoint = InvoicesEndpoint()
plans_endpoint = PlansEndpoint()
subscriptions_endpoint = SubscriptionsEndpoint()
endpoints = [accounts_endpoint,
             adjustments_endpoint,
             transactions_endpoint,
             coupons_endpoint,
             invoices_endpoint,
             plans_endpoint,
             subscriptions_endpoint]


def clear_endpoints():
    """Clear state off of all endpoints. This ensures that no residual state
    carries over between mocurly contexts.
    """
    transactions_endpoint.clear_state()



"""Utility functions that help the development
"""
import pytz
import datetime
from xml.dom import minidom

from jinja2 import Environment, PackageLoader
jinja2_env = Environment(loader=PackageLoader('mocurly', 'templates'), extensions=['jinja2.ext.with_'])


def current_time():
    """Returns the current time in UTC, with the timezone set
    """
    return datetime.datetime.utcnow().replace(tzinfo=pytz.utc)


def details_route(method, uri, is_list=False):
    """A decorator for Endpoint classes to define a custom URI.

    Extends the endpoint's details route. For example, suppose the following
    method was defined on an endpoint with the base uri as `foo`:

    ::

        @details_route('GET', 'bar')
        def bar_callback(self, pk):
            return response

    This will generate a new endpoint /foo/:pk/bar that routes to the
    `bar_callback` method.
    """
    def details_route_decorator(func):
        func.is_route = True
        func.method = method
        func.uri = uri
        func.is_list = is_list
        return func
    return details_route_decorator


def serialize_list(template, object_type_plural, object_type, object_list):
    """Serializes a list of resource objects into its XML version.

    Accepts:
        template - The jinja2 template to use to generate the XML
        object_type_plural - Plural notation for the object type
        object_type - String representation of the object type
        object_list - The list of object to be serialized

    Returns:
        An XML string representing the serialized object list
    """
    serialized_obj_list = []
    for obj in object_list:
        serialized_obj_list.append(serialize(template, object_type, obj))
    return '<{0} type="array">{1}</{0}>'.format(object_type_plural, ''.join(serialized_obj_list)), len(serialized_obj_list)


def serialize(template, object_type, object_dict):
    """Serializes a resource object into its XML version.

    Accepts:
        template - The jinja2 template to use to generate the XML
        object_type - String representation of the object type
        object_dict - The object to be serialized

    Returns:
        An XML string representing the serialized object list
    """
    template = jinja2_env.get_template(template)
    kwargs = {}
    kwargs[object_type] = object_dict
    return template.render(**kwargs)


def deserialize(xml):
    """Deserialize the XML string into an object

    Accepts:
        xml - XML string representing a resource object

    Returns:
        Tuple of object_type and the object as a dictionary. The type can be
        used to resolve what endpoint or object store to route to.
    """
    parsed_xml = minidom.parseString(xml)
    # only know of xml input with 1 child node, since thats all there is to
    # recurly
    assert len(parsed_xml.childNodes) == 1
    root = parsed_xml.firstChild
    if root.hasAttribute('type') and root.getAttribute('type') == 'array':
        return _deserialize_list(root)
    else:
        return _deserialize_item(root)


def _deserialize_list(root):
    """Deserializes a list of objects into their dictionary form.
    """
    return [_deserialize_item(node)[1] for node in root.childNodes]


def _deserialize_item(root):
    object_type = root.tagName
    obj = {}
    for node in root.childNodes:
        if node.hasAttribute('nil'):
            obj[node.tagName] = None
        elif len(node.childNodes) == 1 and node.childNodes[0].nodeType == minidom.Node.TEXT_NODE:
            obj[node.tagName] = node.firstChild.nodeValue
        elif node.hasAttribute('type') and node.getAttribute('type') == 'array':
            obj[node.tagName] = _deserialize_list(node)
        else:
            child_object_type, child_object = _deserialize_item(node)
            obj[node.tagName] = child_object
    return object_type, obj



"""Defines errors that can be returned from recurly
"""

# Error codes used key into messages to be returned. This should mirror that of
# recurly.
TRANSACTION_DECLINED = 'declined'
ERROR_CODES = [TRANSACTION_DECLINED]

# Represents error objects that are used to generate the XML response in case
# of simulated failure
# Requires:
#   error_code - Recurly error code that represnts the particular error
#   error_category - Category of the error
#   customer - Error message that should be used for the customer, as provided
#       by recurly.
#   merchant - Error message that should be used for the merchant, as provided
#       by recurly.
TRANSACTION_ERRORS = {}
TRANSACTION_ERRORS[TRANSACTION_DECLINED] = {
    'error_code': TRANSACTION_DECLINED,
    'error_category': 'declined',
    'customer': 'The transaction was declined. Please use a different card or contact your bank.',
    'merchant': 'The transaction was declined without specific information.  Please contact your payment gateway for more details or ask the customer to contact their bank.'
}


class ResponseError(Exception):
    """Exception class used to signal returning an error response from the
    mocked endpoints

    Should only be raised from within the HTTPretty request callbacks.
    """
    def __init__(self, status_code, response_body):
        self.status_code = status_code
        self.response_body = response_body

    def __str__(self):
        return repr(self.status_code)



from setuptools import setup, find_packages

install_requires = open('requirements.txt').read().split()

setup(
    name='mocurly',
    packages=find_packages(exclude=("tests", "tests.*")),
    package_data={'mocurly': ['templates/*.xml']},
    version='0.2.1',
    description='A library that allows your python tests to easily mock out the recurly library',
    author='Yoriyasu Yano',
    author_email='yoriy@captricity.com',
    url='https://github.com/Captricity/mocurly',
    download_url='https://github.com/Captricity/mocurly/tarball/v0.2.1',
    keywords = ['testing'],
    install_requires=install_requires,
    test_suite='tests'
)



"""Core functions used to interface into mocurly

Exposes the main mocurly class which is the gateway into setting up the mocurly
context.
"""
import recurly
import re
import ssl
import functools
from six.moves.urllib.parse import urlparse, parse_qs
from httpretty import HTTPretty

from .utils import deserialize
from .errors import ResponseError
from .backend import clear_backends


class mocurly(object):
    """Main class that provides the mocked context.

    This can be used as a decorator, as a context manager, or manually. In all
    three cases, the guarded context will route all recurly requests to the
    mocked callback functions defined in endpoints.py.
    """
    def __init__(self, func=None):
        self.started = False
        HTTPretty.reset()

        self.timeout_filter = None
        self.timeout_connection = False
        self.timeout_connection_successful_post = False
        self.func = func

    def __call__(self, *args, **kwargs):
        self.start()
        try:
            retval = self.func(*args, **kwargs)
        finally:
            self.stop()
        return retval

    def __enter__(self):
        self.start()

    def __exit__(self, type, value, tb):
        self.stop()

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)

    def start(self):
        """Starts the mocked context by enabling HTTPretty to route requests to
        the defined endpoints.
        """
        from .endpoints import clear_endpoints
        self.started = True
        clear_endpoints()
        clear_backends()

        if not HTTPretty.is_enabled():
            HTTPretty.enable()

        self._register()

    def stop(self):
        """Stops the mocked context, restoring the routes back to what they were
        """
        if not self.started:
            raise RuntimeError('Called stop() before start()')

        HTTPretty.disable()

    def start_timeout(self, timeout_filter=None):
        """Notifies mocurly to start simulating time outs within the current
        context.

        You can pass in a filter function which will be used to decide what
        requests to timeout. The function will get one parameter, `request`,
        which is an instance of the HTTPrettyRequest class, and should return a
        boolean which when True, the request will time out.

        The following attributes are available on the request object:
            `headers` -> a mimetype object that can be cast into a dictionary,
            contains all the request headers.

            `method` -> the HTTP method used in this request.

            `path` -> the full path to the requested URI.

            `querystring` -> a dictionary containing lists with the attributes.

            `body` -> the raw contents of the request body.

            `parsed_body` -> a dictionary containing parsed request body or
                `None` if `HTTPrettyRequest` doesn't know how to parse it.
        """
        self.timeout_filter = timeout_filter
        self.timeout_connection = True

    def should_timeout(self, request):
        if self.timeout_filter is None or self.timeout_filter(request):
            return self.timeout_connection
        return False

    def stop_timeout(self):
        """Notifies mocurly to stop simulating time outs within the current
        context.
        """
        self.timeout_filter = None
        self.timeout_connection = False

    def start_timeout_successful_post(self, timeout_filter=None):
        """Notifies mocurly to make timeouts on POST requests, but only after
        it has caused state changes.

        Like `start_timeout`, you can pass in a filter function used to decide
        which requests to cause the timeout on.
        """
        self.timeout_filter = timeout_filter
        self.timeout_connection_successful_post = True

    def should_timeout_successful_post(self, request):
        # only applies to POST requests
        if request.method != 'POST':
            return False
        if self.timeout_filter is None or self.timeout_filter(request):
            return self.timeout_connection_successful_post
        return False

    def stop_timeout_successful_post(self):
        """Notifies mocurly to stop simulating successful POST time outs within
        the current context.
        """
        self.timeout_filter = None
        self.timeout_connection_successful_post = False

    def register_transaction_failure(self, account_code, error_code):
        """Register a transaction failure for the given account.

        This will setup mocurly such that all transactions made by the account
        with the given `account_code` will fail with the selected `error_code`.
        """
        from .endpoints import transactions_endpoint
        transactions_endpoint.register_transaction_failure(account_code, error_code)

    def _register(self):
        """Walks the endpoints to register all its URIs to HTTPretty so that
        they can mock recurly requests.
        """
        from .endpoints import endpoints
        for endpoint in endpoints:
            # register list views
            list_uri = recurly.base_uri() + endpoint.base_uri

            def list_callback(request, uri, headers, endpoint=endpoint):
                xml, item_count = endpoint.list()
                headers['X-Records'] = item_count
                return 200, headers, xml
            HTTPretty.register_uri(HTTPretty.GET, list_uri, body=_callback(self)(list_callback), content_type="application/xml")

            def create_callback(request, uri, headers, endpoint=endpoint):
                return 200, headers, endpoint.create(deserialize(request.body)[1])
            HTTPretty.register_uri(HTTPretty.POST, list_uri, body=_callback(self)(create_callback), content_type="application/xml")

            # register details views
            detail_uri = recurly.base_uri() + endpoint.base_uri + r'/([^/ ]+)'
            detail_uri_re = re.compile(detail_uri + r'$')

            def retrieve_callback(request, uri, headers, endpoint=endpoint, detail_uri_re=detail_uri_re):
                pk = detail_uri_re.match(uri).group(1)
                return 200, headers, endpoint.retrieve(pk)
            HTTPretty.register_uri(HTTPretty.GET, detail_uri_re, body=_callback(self)(retrieve_callback), content_type="application/xml")

            def update_callback(request, uri, headers, endpoint=endpoint, detail_uri_re=detail_uri_re):
                pk = detail_uri_re.match(uri).group(1)
                return 200, headers, endpoint.update(pk, deserialize(request.body)[1])
            HTTPretty.register_uri(HTTPretty.PUT, detail_uri_re, body=_callback(self)(update_callback), content_type="application/xml")

            def delete_callback(request, uri, headers, endpoint=endpoint, detail_uri_re=detail_uri_re):
                parsed_url = urlparse(uri)
                pk = detail_uri_re.match('{0}://{1}{2}'.format(parsed_url.scheme, parsed_url.netloc, parsed_url.path)).group(1)
                endpoint.delete(pk, **parse_qs(parsed_url.query))
                return 204, headers, ''
            HTTPretty.register_uri(HTTPretty.DELETE, detail_uri_re, body=_callback(self)(delete_callback))

            # register extra views
            for method in filter(lambda method: callable(method) and getattr(method, 'is_route', False), (getattr(endpoint, m) for m in dir(endpoint))):
                uri = detail_uri + '/' + method.uri
                uri_re = re.compile(uri)

                def extra_route_callback(request, uri, headers, method=method, uri_re=uri_re):
                    pk = uri_re.match(uri).group(1)
                    if method.method == 'DELETE':
                        status = 204
                    else:
                        status = 200
                    if request.method in ['POST', 'PUT']:
                        post_data = request.querystring.copy()
                        if request.body:
                            post_data.update(deserialize(request.body)[1])
                        result = method(pk, post_data)
                    elif method.is_list:
                        result = method(pk, filters=request.querystring)
                        headers['X-Records'] = result[1]
                        result = result[0]
                    else:
                        result = method(pk)
                    return status, headers, result
                if method.method == 'DELETE':
                    HTTPretty.register_uri(HTTPretty.DELETE, uri_re, body=_callback(self)(extra_route_callback))
                else:
                    HTTPretty.register_uri(method.method, uri_re, body=_callback(self)(extra_route_callback), content_type="application/xml")


class _callback(object):
    """Decorator for setting up callback functions to be used in the mocurly
    context.

    This will handle the machinery behind timeout and error simulation.
    """
    def __init__(self, mocurly_instance):
        self.mocurly_instance = mocurly_instance

    def __call__(self, func):
        def wrapped(request, uri, headers, **kwargs):
            # If we want to timeout the request, timeout, but only if we aren't
            # going to allow the POST
            if (self.mocurly_instance.should_timeout(request) and
                    not self.mocurly_instance.should_timeout_successful_post(request)):
                raise ssl.SSLError('The read operation timed out')

            try:
                return_val = func(request, uri, headers, **kwargs)
            except ResponseError as exc:
                # Pass through response errors in a way that httpretty will
                # respond with the right status code and message
                if not self.mocurly_instance.should_timeout_successful_post(request):
                    return exc.status_code, headers, exc.response_body

            if self.mocurly_instance.should_timeout_successful_post(request):
                raise ssl.SSLError('The read operation timed out')

            return return_val
        return wrapped



"""In-memory database backends for each recurly resource
"""
import six


class BaseBackend(object):
    """Datastore to store resource objects in memory throughout the recurly context.
    """
    def __init__(self):
        self.datastore = {}

    def empty(self):
        """Whether or not the datastore is empty
        """
        return not bool(self.datastore)

    def has_object(self, uuid):
        """Whether or not the datastore has an object with the requested id
        """
        return uuid in self.datastore

    def add_object(self, uuid, obj):
        """Add the provided object into the datastore
        """
        self.datastore[uuid] = obj.copy()
        return obj

    def list_objects(self, filter_pred=lambda x: True):
        """List the objects in the datastore.

        You can pass in a filter function that returns a boolean given a
        resource object to limit the number of objects to return.
        """
        return list(six.moves.filter(filter_pred, [v.copy() for v in self.datastore.values()]))

    def get_object(self, uuid):
        """Retrieve the object with the given id from the datastore
        """
        return self.datastore[uuid].copy()

    def update_object(self, uuid, updated_data):
        """Update the object with the given id with the new information
        """
        obj = self.datastore[uuid]
        obj.update(updated_data)
        return obj.copy()

    def delete_object(self, uuid):
        """Delete the object with the given id from the datastore
        """
        del self.datastore[uuid]

    def clear_all(self):
        """Clear all objects from the datastore
        """
        self.datastore = {}


class AccountBackend(BaseBackend):
    pass


class BillingInfoBackend(BaseBackend):
    pass


class InvoiceBackend(BaseBackend):
    pass


class CouponBackend(BaseBackend):
    pass


class CouponRedemptionBackend(BaseBackend):
    pass


class PlanBackend(BaseBackend):
    pass


class PlanAddOnBackend(BaseBackend):
    pass


class SubscriptionBackend(BaseBackend):
    pass


class TransactionBackend(BaseBackend):
    pass


class AdjustmentBackend(BaseBackend):
    pass


# Provide public access to each resource backend, so that users can do low
# level object checking in their tests (e.g query objects that were created as
# a side effect of an action)
accounts_backend = AccountBackend()
billing_info_backend = BillingInfoBackend()
invoices_backend = InvoiceBackend()
coupons_backend = CouponBackend()
coupon_redemptions_backend = CouponRedemptionBackend()
plans_backend = PlanBackend()
plan_add_ons_backend = PlanAddOnBackend()
subscriptions_backend = SubscriptionBackend()
transactions_backend = TransactionBackend()
adjustments_backend = AdjustmentBackend()


def clear_backends():
    """Clears all resource datastores. This ensures that no residual state
    carries over across mocurly contexts.
    """
    accounts_backend.clear_all()
    billing_info_backend.clear_all()
    invoices_backend.clear_all()
    coupons_backend.clear_all()
    coupon_redemptions_backend.clear_all()
    plans_backend.clear_all()
    plan_add_ons_backend.clear_all()
    subscriptions_backend.clear_all()
    transactions_backend.clear_all()
    adjustments_backend.clear_all()
